#include "config.h"

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include "getopt.h"
#include "tcpclient.h"
#include "frontend.h"
#include "cursesfront.h"
#include "panelfront.h"
#include "configfile.h"

#define MAJOR 2
#define MINOR 0

enum gui_type {
    gui_ncurses, gui_panel
};

const char *config_base = "kismet.conf";

TcpClient kismet_serv;
Frontend *gui;
gui_type guitype;
char *configfile;
char *server = NULL;
int sound = -1;
int speech = -1;

int group_track = 0;
string configdir, groupfile;
FILE *group_file = NULL;

// Catch our interrupt
void CatchShutdown(int sig) {

    if (group_track) {
        if ((group_file = fopen(groupfile.c_str(), "w")) == NULL) {
            fprintf(stderr, "WARNING: Unable to open '%s' for writing, groups will not be saved.\n",
                    groupfile.c_str());
        } else {
            gui->WriteGroupMap(group_file);
        }
    }

    if (gui != NULL && sig != SIGHUP)
        gui->EndDisplay();

    exit(0);
}

int Usage(char *argv) {
    printf("Usage: %s [OPTION]\n", argv);
    printf("Most (or all) of these options can (and should) be configured via the\n"
           "kismet.conf global config file, but can be overridden here.\n");
    printf(
           "  -f, --config-file <file>     Use alternate config file\n"
           "  -q, --quiet                  Don't play sounds\n"
           "  -s, --server <host:port>     Connect to Kismet host and port\n"
           "  -g, --gui                    GUI type to create (curses, panel)\n"
           "  -c, --columns                Columns to display initially (comma seperated)\n"
           "  -v, --version                Kismet version\n"
           "  -h, --help                   What do you think you're reading?\n");
    exit(1);
}


// Fork and run a system call to play a sound
void PlaySound(string player, string sound, map<string, string> soundmap) {
    // Why doesn't the above work?  It seems to bugger up wtap dumping.
    // If we don't have this sound event defined
    if (soundmap.find(sound) == soundmap.end())
        return;

    char snd_call[1024];
    snprintf(snd_call, 1024, "%s %s >/dev/null 2>/dev/null &", player.c_str(),
             soundmap[sound].c_str());
    if (system(snd_call) < 0) {
        char status[STATUS_MAX];
        snprintf(status, STATUS_MAX, "ERROR:  Could not execute system() for `%s'",
                 snd_call);
        gui->WriteStatus(status);
    }

}

// Fork and run a system call to play a sound
void SayText(string player, string text) {
    char snd_call[1024];

    snprintf(snd_call, 1024, "echo '(SayText \"%s\")' | %s >/dev/null 2>/dev/null &", text.c_str(),
             player.c_str());

    if (system(snd_call) < 0) {
        char status[STATUS_MAX];
        snprintf(status, STATUS_MAX, "ERROR:  Could not execute system() for `%s'",
                 snd_call);
        gui->WriteStatus(status);
    }
}

char *exec_name;
int main(int argc, char *argv[]) {
    exec_name = argv[0];

    int sleepu = 0;
    time_t last_draw = time(0);
    char status[STATUS_MAX];

    const char *reqgui = NULL;

    const char *sndplay = NULL;
    const char *festival = NULL;

    const char *columns = NULL;

    map<string, string> wav_map;

    char guihost[1024];
    int guiport = -1;

    int gpsmode = -1;

    static struct option long_options[] = {   /* options table */
        { "config-file", required_argument, 0, 'f' },
        { "quiet", no_argument, 0, 'q' },
        { "server", required_argument, 0, 's' },
        { "help", no_argument, 0, 'h' },
        { "version", no_argument, 0, 'v' },
        { "columns", required_argument, 0, 'c' },
        { 0, 0, 0, 0 }
    };
    int option_index;
    int decay = 5;

    // Catch the interrupt handler to shut down
    signal(SIGINT, CatchShutdown);
    signal(SIGTERM, CatchShutdown);
    signal(SIGHUP, CatchShutdown);
    signal(SIGPIPE, SIG_IGN);

    while(1) {
        int r = getopt_long(argc, argv, "f:qs:hvg:",
                            long_options, &option_index);
        if (r < 0) break;
        switch(r) {
        case 'f':
            // Config path
            configfile = optarg;
            fprintf(stderr, "Using alternate config file: %s\n", configfile);
            break;
        case 'q':
            sound = 0;
            break;
        case 'g':
            reqgui = optarg;
            break;
        case 'v':
            fprintf(stderr, "Kismet curses %d.%d\n", MAJOR, MINOR);
            break;
        case 's':
            server = optarg;
            fprintf(stderr, "Using server: %s\n", server);
            break;
        case 'c':
            columns = optarg;
            break;
        default:
            Usage(argv[0]);
            break;
        }
    }

    // If we haven't gotten a command line config option...
    if (configfile == NULL) {
        configfile = (char *) malloc(1024*sizeof(char));
        snprintf(configfile, 1024, "%s/%s", SYSCONF_LOC, config_base);
    }

    // Parse the config and load all the values from it and/or our command
    // line options.  This is a little soupy but it does the trick.
    if (ParseConfig(configfile) < 0) {
        exit(1);
    }

    if (FetchOpt("configdir") != "") {
        configdir = ExpandLogPath(FetchOpt("configdir"), "", "", 0, 1);
    } else {
        fprintf(stderr, "FATAL:  No 'configdir' option in the config file.\n");
        exit(1);
    }

    if (FetchOpt("groupmap") != "") {
        // Explode the group map path
        groupfile = ExpandLogPath(configdir + FetchOpt("groupmap"), "", "", 0, 1);
        group_track = 1;
    }

    if (reqgui == NULL) {
        if (FetchOpt("gui") == "") {
            fprintf(stderr, "ERROR:  No gui given in config file and none given on the command line.\n");
            exit(1);
        }

        reqgui = FetchOpt("gui").c_str();
    }

    if (!strcasecmp(reqgui, "curses")) {
#if defined(HAVE_LIBNCURSES) && defined(BUILD_CURSES)
        gui = new NCurseFront;
        guitype = gui_ncurses;
#else
        fprintf(stderr, "ERROR:  Curses support not compiled in.\n");
        exit(1);
#endif
    } else if (!strcasecmp(reqgui, "panel")) {
#if defined(HAVE_LIBPANEL) && defined(BUILD_PANEL) && defined(HAVE_LIBNCURSES)
        gui = new PanelFront;
        guitype = gui_panel;
#else
        fprintf(stderr, "ERROR:  Panels support not compiled in.\n");
        exit(1);
#endif
    } else {
        fprintf(stderr, "ERROR:  Unknown GUI type requested ('%s')\n", reqgui);
        exit(1);
    }

    if (columns == NULL) {
        if (FetchOpt("guicolumns") == "") {
            fprintf(stderr, "FATAL:  No columns in the config file and none given on the command line.\n");
            exit(1);
        }

        columns = FetchOpt("guicolumns").c_str();
    }

    if (server == NULL) {
        server = (char *) FetchOpt("guihost").c_str();
    }

    if (FetchOpt("guisound") == "true" && sound == -1) {
        if (FetchOpt("guisoundplay") != "") {
            sndplay = FetchOpt("guisoundplay").c_str();
            sound = 1;

            if (FetchOpt("guisound_new") != "")
                wav_map["new"] = FetchOpt("guisound_new");
            if (FetchOpt("guisound_traffic") != "")
                wav_map["traffic"] = FetchOpt("guisound_traffic");
            if (FetchOpt("guisound_junktraffic") != "")
                wav_map["junktraffic"] = FetchOpt("guisound_junktraffic");
            if (FetchOpt("guisound_gpslock") != "")
                wav_map["gpslock"] = FetchOpt("guisound_gpslock");
            if (FetchOpt("guisound_gpslost") != "")
                wav_map["gpslost"] = FetchOpt("guisound_gpslost");

        } else {
            fprintf(stderr, "ERROR:  Sound alerts enabled but no sound playing binary specified.\n");
        }
    } else if (sound == -1)
        sound = 0;

    /* Added by Shaw Innes 17/2/02 */
    if (FetchOpt("guispeech") == "true" && speech == -1) {
        if (FetchOpt("guifestival") != "") {
            festival = FetchOpt("guifestival").c_str();
            speech = 1;
        } else {
            fprintf(stderr, "ERROR: Speech alerts enabled but no path to festival has been specified.\n");
        }
    } else if (speech == -1)
        speech = 0;

    if (FetchOpt("guidecay") != "") {
        if (sscanf(FetchOpt("guidecay").c_str(), "%d", &decay) != 1) {
            fprintf(stderr, "FATAL:  Illegal config file value for guidecay.\n");
            exit(1);
        }
    }

    if (FetchOpt("microsleep") == "") {
        sleepu = 100;
    } else if (sscanf(FetchOpt("microsleep").c_str(), "%d", &sleepu) != 1) {
        fprintf(stderr, "FATAL:  Illegal config file value for microsleep.\n");
        exit(1);
    }

    if (sscanf(server, "%[^:]:%d", guihost, &guiport) != 2) {
        fprintf(stderr, "FATAL:  Invalid server (%s) specified (host:port required)\n",
               server);
        exit(1);
    }

    // handle the config bits
    struct stat fstat;
    if (stat(configdir.c_str(), &fstat) == -1) {
        fprintf(stderr, "NOTICE: configdir '%s' does not exist, making it.\n",
                configdir.c_str());
        if (mkdir(configdir.c_str(), S_IRUSR | S_IWUSR | S_IXUSR) < 0) {
            fprintf(stderr, "FATAL:  Could not make configdir: %s\n",
                    strerror(errno));
            exit(1);
        }
    } else if (! S_ISDIR(fstat.st_mode)) {
        fprintf(stderr, "FATAL: configdir '%s' exists but is not a directory.\n",
                configdir.c_str());
        exit(1);
    }

    if (group_track) {
        if (stat(groupfile.c_str(), &fstat) == -1) {
            fprintf(stderr, "NOTICE:  Group file did not exist, it will be created.\n");
        } else {
            if ((group_file = fopen(groupfile.c_str(), "r")) == NULL) {
                fprintf(stderr, "FATAL:  Could not open group file '%s': %s\n",
                        groupfile.c_str(), strerror(errno));
                exit(1);
            }

            gui->ReadGroupMap(group_file);

            fclose(group_file);
        }
    }

    if (kismet_serv.Connect(guiport, guihost) < 0) {
        fprintf(stderr, "FATAL:  Could not connect to %s:%d.\n", guihost, guiport);
        exit(1);
    }

    time_t serv_start = 0;
    // Spin for 15 seconds until we get a header from the server, or die
    fprintf(stderr, "Looking for startup info from %s:%d...", guihost, guiport);
    int header_count = 0;

    while (serv_start == 0) {

        fprintf(stderr, ".");

        kismet_serv.Poll();

        if (header_count++ >= 20) {
            fprintf(stderr, " failed.\nFATAL:  Did not get startup info from %s:%d within 20 seconds.\n",
                    guihost, guiport);
            exit(1);
        }

        serv_start = kismet_serv.FetchStart();

        sleep(1);
    }

    fprintf(stderr, " found.\nConnected to Kismet server %d.%d on %s:%d\n",
            kismet_serv.FetchMajor(), kismet_serv.FetchMinor(), guihost, guiport);

#if defined(HAVE_LIBPANEL) && defined(HAVE_LIBNCURSES) && defined(BUILD_PANEL)
    // I don't like how this is done
    if (guitype == gui_panel) {
        PanelFront *pgui = (PanelFront *) gui;
        pgui->SetColumns(columns);
    }
#endif

    gui->AddClient(&kismet_serv);

    gui->InitDisplay(decay, serv_start);

    snprintf(status, STATUS_MAX, "Connected to Kismet server version %d.%d on %s:%d",
             kismet_serv.FetchMajor(), kismet_serv.FetchMinor(), guihost, guiport);
    gui->WriteStatus(status);

    int num_networks = 0, num_packets = 0, num_noise = 0, num_dropped = 0;

    time_t last_click = time(0);
    timeval ts;
    while (1) {
        timeval new_ts;
        gettimeofday(&new_ts, NULL);

        if (ts.tv_sec != new_ts.tv_sec || new_ts.tv_usec - ts.tv_usec < 200000) {
            int poll_ret;
            poll_ret = gui->Poll();

            if (poll_ret == FE_QUIT)
                CatchShutdown(-1);
        }
        ts = new_ts;


        if (kismet_serv.Poll() < 0) {
            snprintf(status, 80, "TCP server error: %s", kismet_serv.FetchError());
            gui->WriteStatus(status);
        }

        if (strlen(kismet_serv.FetchStatus()) != 0) {
            gui->WriteStatus(kismet_serv.FetchStatus());
            gui->DrawDisplay();
        }

        if (kismet_serv.FetchMode() == 0 && gpsmode != 0) {
            if (sound == 1 && gpsmode != -1)
                PlaySound(sndplay, "gpslost", wav_map);
            gpsmode = 0;
        } else if (kismet_serv.FetchMode() != 0 && gpsmode == 0) {
            if (sound == 1 && gpsmode != -1)
                PlaySound(sndplay, "gpslock", wav_map);
            gpsmode = 1;
        }


        if (kismet_serv.FetchNumNetworks() != num_networks) {
            if (sound == 1)
                PlaySound(sndplay, "new", wav_map);
        }
        num_networks = kismet_serv.FetchNumNetworks();

        if (kismet_serv.FetchNumNetworks() != num_networks && speech == 1) {
                char text[100];

                wireless_network *newnet = kismet_serv.FetchNthRecent(1)[0];

                snprintf(text, 100, "New %s network '%s' detected.",
                         (newnet->wep ? "En-crypted" : "Un-en-crypted"),
                         newnet->ssid.c_str());

                SayText(festival, text);
        }


        if (kismet_serv.FetchNumPackets() != num_packets ) {
            if (time(0) - last_click >= decay && sound == 1) {

                if (kismet_serv.FetchNumPackets() - num_packets >
                    kismet_serv.FetchNumDropped() - num_dropped) {
                    PlaySound(sndplay, "traffic", wav_map);
                } else {
                    PlaySound(sndplay, "junktraffic", wav_map);
                }
                
                last_click = time(0);
            }

            num_packets = kismet_serv.FetchNumPackets();
            num_noise = kismet_serv.FetchNumNoise();
            num_dropped = kismet_serv.FetchNumDropped();
        }

        // Draw if it's time
        if (time(0) != last_draw) {
            gui->DrawDisplay();
            last_draw = time(0);
        }

        if (sleepu > 0)
            usleep(sleepu);
    }
}

