#ifndef __PACKETRACKER_H__
#define __PACKETRACKER_H__

#include "config.h"

#include <stdio.h>
#include <time.h>
#include <list>
#include <map>
#include <vector>
#include <algorithm>
#include <string>

#include "gpsd.h"
#include "packet.h"

// Extern over to kismet_server.cc to get our first time
extern time_t start_time;

#define MAC_STR_LEN (MAC_LEN * 2) + 6

const string NOSSID("<no ssid>");

enum wireless_network_type {
    network_ap,
    network_adhoc,
    network_probe,
    network_data
};

enum address_type {
    address_none,
    address_udp,
    address_arp,
    address_dhcp,
    address_group
};

// A network
typedef struct {
    wireless_network_type type;

    string ssid;

    string beacon_info;

    // Packet counts
    int llc_packets;
    int data_packets;
    int crypt_packets;
    int interesting_packets;

    // info extracted from packets
    //uint8_t bssid[MAC_LEN];
    int channel;
    int wep;
    string bssid;

    uint8_t bssid_raw[MAC_LEN];

    // Are we a cloaked SSID?
    int cloaked;

    // Last time we saw a packet
    time_t last_time;
    // First packet
    time_t first_time;

    // beacon interval
    int beacon;

    // Are we in the list?
    int listed;

    address_type atype;

    // How many address octets have we matched?
    int octets;

    // What IP range are we using?
    uint8_t range_ip[4];
    // What netmask are we using?
    uint8_t mask[4];
    // What's the gateway?
    uint8_t gate_ip[4];

    map<string, cdp_packet> cisco_equip;

    /*
    float gps_lat, gps_lon, gps_alt, gps_spd;
    int gps_mode;

    float first_lat, first_lon, first_alt, first_spd;
    int first_mode;
    */

    int gps_fixed;
    float min_lat, min_lon, min_alt, min_spd;
    float max_lat, max_lon, max_alt, max_spd;

    // How fast we can go
    double maxrate;

} wireless_network;

enum host_type {
    host_unknown, host_linux, host_windows, host_mac
};

typedef struct {
    string mac;

    host_type type;

} wireless_host;

class Packetracker {
public:
    Packetracker();

    // Get the error
    char *FetchError() { return errstr; }

    // Set up the gps
    void AddGPS(GPSD *in_gps);

    // Set up filtering
    void AddFilter(string in_filter) { filter = in_filter; }

    // Packet tracker stuff
    int ProcessPacket(packet_info info, char *in_status);

    // Get all the networks
    vector<wireless_network *> FetchNetworks();

    int WriteNetworks(FILE *in_file);
    int WriteCSVNetworks(FILE *in_file);
    int WriteXMLNetworks(FILE *in_file);
    int WriteCisco(FILE *in_file);

    void WriteSSIDMap(FILE *in_file);
    void ReadSSIDMap(FILE *in_file);

    // Convert the MAC
    static string Mac2String(uint8_t *mac, char seperator);
    // Utility to find strings that are empty or contain all spaces
    static bool IsBlank(const char *s);

    int FetchNumNetworks() { return num_networks; }
    int FetchNumPackets() { return num_packets; }
    int FetchNumDropped() { return num_dropped; }
    int FetchNumNoise() { return num_noise; }
    int FetchNumCrypt() { return num_crypt; }
    int FetchNumInteresting() { return num_interesting; }
    int FetchNumCisco() { return num_cisco; }

    static string Net2String(wireless_network *in_net);
    static string CDP2String(cdp_packet *in_cdp);

protected:
    char errstr[1024];
    GPSD *gps;

    int num_networks, num_packets, num_dropped, num_noise,
        num_crypt, num_interesting, num_cisco;

    // List of MAC's to filter
    string filter;

    // all the networks
    vector<wireless_network *> network_list;

    // Several maps to refer to networks
    map<string, wireless_network *> ssid_map;
    map<string, wireless_network *> bssid_map;
    // client MAC to BSSID
    map<string, string> client_map;

    // Map BSSID's to SSID for storage and cloaking
    map<string, string> bssid_cloak_map;
};

#endif

