#include "tcpclient.h"
#include "networksort.h"

TcpClient::TcpClient() {
    sv_valid = 0;
    client_fd = 0;

    lat = lon = alt = spd = 0;
    mode = 0;

    num_networks = num_packets = num_crypt = num_interesting =
        num_noise = num_dropped = 0;

    start_time = 0;
    major = minor = 0;

    power = quality = noise = 0;

    memset(status, 0, STATUS_MAX);
}

TcpClient::~TcpClient() {
    if (sv_valid)
        close(client_fd);

    sv_valid = 0;
}

int TcpClient::Connect(short int in_port, char *in_host) {
    // Copy the port to our local data
    port = in_port;

    // Resolve the hostname we were given/found to see if it's actually
    // valid
    if ((client_host = gethostbyname(in_host)) == NULL) {
        snprintf(errstr, 1024, "TcpClient could not resolve host \"%s\"\n", in_host);
        return (-1);
    }

    strncpy(hostname, in_host, MAXHOSTNAMELEN);

    // Set up our socket
    bzero(&client_sock, sizeof(client_sock));
    client_sock.sin_family = client_host->h_addrtype;
    memcpy((char *) &client_sock.sin_addr.s_addr, client_host->h_addr_list[0],
           client_host->h_length);
    client_sock.sin_port = htons(in_port);


    // Debug("Server::Setup calling socket()");
    if ((client_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
        snprintf(errstr, 1024, "TcpClient socket() failed %d (%s)\n",
                 errno, sys_errlist[errno]);
        return (-2);
    }

    // Bind to the local half of the pair
    local_sock.sin_family = AF_INET;
    local_sock.sin_addr.s_addr = htonl(INADDR_ANY);
    local_sock.sin_port = htons(0);

    if (bind(client_fd, (struct sockaddr *) &local_sock, sizeof(local_sock)) < 0) {
        snprintf(errstr, 1024, "FATAL: TcpClient bind() failed %d (%s)\n",
                 errno, sys_errlist[errno]);
        return (-3);
    }

    // Connect
    if (connect(client_fd, (struct sockaddr *) &client_sock, sizeof(client_sock)) < 0) {
        snprintf(errstr, 1024, "FATAL: TcpClient connect() failed %d (%s)\n",
                 errno, sys_errlist[errno]);
        return (-4);
    }

    int save_mode = fcntl(client_fd, F_GETFL, 0);
    fcntl(client_fd, F_SETFL, save_mode | O_NONBLOCK);

    sv_valid = 1;

    clientf = fdopen(client_fd, "r");

    return 1;
}

int TcpClient::Poll() {
    // Clear the status
    memset(status, 0, STATUS_MAX);

    int selected;
    fd_set read_set;
    fd_set except_set;

    FD_ZERO(&read_set);
    FD_SET(client_fd, &read_set);
    FD_ZERO(&except_set);

    struct timeval tm;

    tm.tv_sec = 0;
    tm.tv_usec = 0;

    // Enter the select loop
    //Debug("Server::Poll() - Calling select()");
    if ((selected = select(client_fd+1, &read_set, NULL, &except_set, &tm)) < 0) {
        if (errno != EINTR) {
            snprintf(errstr, 1024, "TcpServer select() returned %d (%s)\n",
                    errno, sys_errlist[errno]);
            return (-1);
        }
    }

    if (FD_ISSET(client_fd, &except_set)) {
        snprintf(errstr, 1024, "Exception on socket.\n");
        return(-1);
    }

    if (!FD_ISSET(client_fd, &read_set))
        return 0;

    char data[2048];
    memset(data, 0, 2048);

    while (1) {
        if (fgets(data, 2048, clientf) == NULL) {
            if (errno != 0 && errno != EAGAIN) {
                snprintf(errstr, 1024, "Read error %d (%s)\n",
                         errno, sys_errlist[errno]);
                return (-1);
            }

            return (0);
        }

        //fprintf(stderr, "Got: '%s'\n", data);

        if (strlen(data) < 2)
            return 0;

        // Drop out now on a status event so we can get drawn
        if (ParseData(data) == 2)
            return 1;
    }

    return 1;
}

int TcpClient::ParseData(char *in_data) {
    char header[64];

    if (sscanf(in_data, "%64[^:]", header) < 1)
        return 0;

    if (!strncmp(header, "*TERMINATE", 64)) {
        sv_valid = 0;
        return 1;
    } else if (!strncmp(header, "*KISMET", 64)) {
        if (sscanf(in_data, "%*[^:]: %d.%d %d",
                   &major, &minor, (int *) &start_time) < 1)
            return 0;
    } else if (!strncmp(header, "*TIME", 64)) {
        if (sscanf(in_data, "%*[^:]: %d", (int *) &serv_time) < 1)
            return 0;

        /*
         printf("Time header %d\n", (int) serv_time);
         */
    } else if (!strncmp(header, "*NETWORK", 64)) {
        wireless_network net;

        int scanned;

        char bssid[18], ssid[256], beacon[256];
        short int range[4], mask[4], gate[4];

        float maxrate;

        scanned = sscanf(in_data, "%*[^:]: %17s %d \001%255[^\001]\001 \001%255[^\001]\001 "
                         "%d %d %d %d %d %d %d %d %d %hd.%hd.%hd.%hd %hd.%hd.%hd.%hd %hd.%hd.%hd.%hd "
                         "%d %f %f %f %f %f %f %f %f %d %d %d %f",
                         bssid, (int *) &net.type, ssid, beacon,
                         &net.llc_packets, &net.data_packets, &net.crypt_packets, &net.interesting_packets,
                         &net.channel, &net.wep, (int *) &net.first_time, (int *) &net.last_time,
                         (int *) &net.atype, &range[0], &range[1], &range[2], &range[3],
                         &mask[0], &mask[1], &mask[2], &mask[3],
                         &gate[0], &gate[1], &gate[2], &gate[3],
                         &net.gps_fixed, &net.min_lat, &net.min_lon, &net.min_alt, &net.min_spd,
                         &net.max_lat, &net.max_lon, &net.max_alt, &net.max_spd,
                         /*
                         &net.gps_lat, &net.gps_lon, &net.gps_alt, &net.gps_spd, &net.gps_mode,
                         &net.first_lat, &net.first_lon, &net.first_alt, &net.first_spd,
                         &net.first_mode,
                         */
                         &net.octets, &net.cloaked, &net.beacon,
                         &maxrate);

        if (scanned < 36)
            return 0;

        // Alignment issues on some platforms make this necessary
        unsigned int rawmac0, rawmac1, rawmac2, rawmac3, rawmac4, rawmac5;

        sscanf(bssid, "%02X:%02X:%02X:%02X:%02X:%02X",
               &rawmac0, &rawmac1, &rawmac2,
               &rawmac3, &rawmac4, &rawmac5);

        net.bssid_raw[0] = rawmac0;
        net.bssid_raw[1] = rawmac1;
        net.bssid_raw[2] = rawmac2;
        net.bssid_raw[3] = rawmac3;
        net.bssid_raw[4] = rawmac4;
        net.bssid_raw[5] = rawmac5;

        /*
        sscanf(bssid, "%02X:%02X:%02X:%02X:%02X:%02X",
               (unsigned int *) &net.bssid_raw[0], (unsigned int *) &net.bssid_raw[1],
               (unsigned int *) &net.bssid_raw[2], (unsigned int *) &net.bssid_raw[3],
               (unsigned int *) &net.bssid_raw[4], (unsigned int *) &net.bssid_raw[5]);
               */

        net.bssid = bssid;
        if (ssid[0] != '\002')
            net.ssid = ssid;
        if (beacon[0] != '\002')
            net.beacon_info = beacon;
        for (int x = 0; x < 4; x++) {
            net.range_ip[x] = (uint8_t) range[x];
            net.mask[x] = (uint8_t) mask[x];
            net.gate_ip[x] = (uint8_t) gate[x];
        }
        net.maxrate = maxrate;

        net_map[net.bssid] = net;

    } else if (!strncmp(header, "*GPS", 64)) {
        if (sscanf(in_data, "%*[^:]: %f %f %f %f %d", &lat, &lon, &alt, &spd, &mode) < 5)
            return 0;

    } else if (!strncmp(header, "*INFO", 64)) {
        if (sscanf(in_data, "%*[^:]: %d %d %d %d %d %d %d %d %d\n",
                   &num_networks, &num_packets,
                   &num_crypt, &num_interesting,
                   &num_noise, &num_dropped, &quality, &power, &noise) < 9)
            return 0;
    } else if (!strncmp(header, "*CISCO", 64)) {
        cdp_packet cdp;
        memset(&cdp, 0, sizeof(cdp_packet));
        char bssid[18];
        int cap0, cap1, cap2, cap3, cap4, cap5, cap6;

        if (sscanf(in_data, "%*[^:]: %s \001%s\001 %d.%d.%d.%d \001%s\001 "
                   "%d:%d:%d:%d;%d;%d;%d \001%s\001 \001%s\001\n",
                   bssid, cdp.dev_id,
                   (int *) &cdp.ip[0], (int *) &cdp.ip[1], (int *) &cdp.ip[2], (int *) &cdp.ip[2],
                   cdp.interface, &cap0, &cap1, &cap2, &cap3, &cap4, &cap5, &cap6,
                   cdp.software, cdp.platform) < 16)
            return 0;

        if (net_map.find(bssid) == net_map.end())
            return 0;

        net_map[bssid].cisco_equip[cdp.dev_id] = cdp;

    } else if (!strncmp(header, "*STATUS", 64)) {
        if (sscanf(in_data, "%*[^:]: %[^\n]\n", status) != 1)
            return 0;
        return 2;
    } else if (!strncmp(header, "*STRING", 64)) {
        char netstr[2048];
        if ((sscanf(in_data, "%*[^:]: %[^\n]\n", netstr)) != 1)
            return 0;

        strings.push_back(netstr);
        if (strings.size() > MAX_STRINGS)
            strings.erase(strings.begin());
    }

    return 1;
}

time_t TcpClient::FetchServTime() {
    return serv_time;
}

int TcpClient::FetchLoc(float *in_lat, float *in_lon, float *in_alt, float *in_spd, int *in_mode) {
    *in_lat = lat; *in_lon = lon;
    *in_alt = alt; *in_spd = spd;
    *in_mode = mode;
    return mode;
}

vector<wireless_network *> TcpClient::FetchNetworkList() {
    vector<wireless_network *> retvec;

    for (map<string, wireless_network>::iterator x = net_map.begin(); x != net_map.end(); ++x) {
        retvec.push_back(&x->second);
    }

    return retvec;
}

vector<wireless_network *> TcpClient::FetchNthRecent(unsigned int n) {
    vector<wireless_network *> vec = FetchNetworkList();

    // XXX
    // This is much easier now that we use vectors.  We're already ordered
    // by time since we're inserted in order, so we can just erase...
    // XXX
    sort(vec.begin(), vec.end(), SortLastTimeLT());

    int drop = vec.size() - n;

    if (drop > 0) {
        vec.erase(vec.begin(), vec.begin() + drop);
    }

    sort(vec.begin(), vec.end(), SortFirstTimeLT());

    return vec;
}

